@@ -5,7 +5,7 @@ use warnings;
 
 use parent qw(Tickit::Widget);
 
-our $VERSION = '0.214';
+our $VERSION = '0.215';
 
 =head1 NAME
 
@@ -13,7 +13,7 @@ Tickit::Widget::Table - table widget with support for scrolling/paging
 
 =head1 VERSION
 
-version 0.214
+version 0.215
 
 =head1 SYNOPSIS
 
@@ -39,8 +39,8 @@ version 0.214
 
 =head1 DESCRIPTION
 
-B<WARNING>: This is a preview release. API is subject to change in future,
-please get in contact if you're using this, or wait for version 1.000.
+B<WARNING>: This is still a preview release. API might be subject to change in
+future, please get in contact if you're using this, or wait for version 1.000.
 
 =begin HTML
 
@@ -154,7 +154,7 @@ such as replacing long strings with an elided version ("Some lengthy messa...")
 use Tickit::RenderBuffer qw(LINE_SINGLE LINE_DOUBLE CAP_BOTH);
 use Tickit::Utils qw(distribute substrwidth align textwidth chars2cols);
 use String::Tagged;
-use Future::Utils qw(fmap_void repeat);
+use Future::Utils qw(fmap_void try_repeat);
 use Tickit::Style;
 use Scalar::Util qw(looks_like_number blessed);
 use POSIX qw(floor);
@@ -794,14 +794,17 @@ sub render_body {
 
 	for my $line ($rect->linerange) {
 		my $idx = $self->idx_from_row($line);
-		my $f = $self->row_cache($idx);
-		if($f->is_done) {
-			$self->render_row($rb, $rect, $idx, $f->get);
-		} elsif($f->is_ready) {
-			$self->render_failed_row($rb, $rect, $idx, $f->is_cancelled ? 'cancelled' : $self->failure_transform($f->failure));
+		if(my $f = $self->row_cache($idx)) {
+			if($f->is_done) {
+				$self->render_row($rb, $rect, $idx, $f->get);
+			} elsif($f->is_ready) {
+				$self->render_failed_row($rb, $rect, $idx, $f->is_cancelled ? 'cancelled' : $self->failure_transform($f->failure));
+			} else {
+				$self->render_pending_row($rb, $rect, $idx);
+				$f->on_done($self->curry::expose_row($idx));
+			}
 		} else {
-			$self->render_pending_row($rb, $rect, $idx);
-			$f->on_done($self->curry::expose_row($idx));
+			$rb->erase_at($line, $rect->left, $rect->cols, $self->get_style_pen);
 		}
 	}
 }
@@ -830,9 +833,12 @@ sub render_row {
 		: undef
 	);
 	for my $col (0..$#$data) {
-		my $v = $self->apply_view_transformations($row, $col, $data->[$col]);
+		my $v = $self->apply_view_transformations($row, $col, $data->[$col]) // '';
 		my $def = $self->{columns}[$col];
 		$rb->goto($line, $def->{start} // 0);
+		# Prevent any vertical whitespace because we only handle single-line widgets, and textwidth() returns
+		# undef on \n
+		$v =~ s/\v+/ /g;
 		my ($pre, undef, $post) = align textwidth($v), $def->{value} // 0, $def->{align} // 0;
 		$rb->erase($pre, $base_pen) if $pre;
 		if(blessed($v) && $v->isa('String::Tagged')) {
@@ -844,10 +850,11 @@ sub render_row {
 			});
 			$st->iter_substr_nooverlap(sub {
 				my ($substr, %tags) = @_;
-				my $pen = Tickit::Pen::Immutable->new(
+				my %attr = (
 					$base_pen->getattrs,
 					%tags
 				);
+				my $pen = Tickit::Pen::Immutable->new(%attr);
 				$rb->text($substr, $pen);
 			});
 		} else {
@@ -871,12 +878,16 @@ sub render_failed_row {
 		: 'failed'
 	);
 
-		$rb->goto($line, 0);
-		my ($pre, undef, $post) = align textwidth($failure), $self->body_cols, 0.5;
-		$rb->erase($pre, $base_pen) if $pre;
-		$rb->text($failure, $base_pen);
-		$rb->erase($post, $base_pen) if $post;
+	$rb->goto($line, 0);
+	($failure //= '') =~ s/\v+/ /g;
+	my $w = textwidth($failure);
+	die "undef \$w from $failure" unless defined $w;
+	my ($pre, undef, $post) = align $w, $self->body_cols // 1, 0.5;
+	$rb->erase($pre, $base_pen) if $pre;
+	$rb->text($failure, $base_pen);
+	$rb->erase($post, $base_pen) if $post;
 }
+
 sub render_pending_row {
 	my ($self, $rb, $rect, $row) = @_;
 
@@ -889,11 +900,11 @@ sub render_pending_row {
 		: 'pending'
 	);
 
-		$rb->goto($line, 0);
-		my ($pre, undef, $post) = align textwidth($self->loading_message), $self->body_cols, 0.5;
-		$rb->erase($pre, $base_pen) if $pre;
-		$rb->text($self->loading_message, $base_pen);
-		$rb->erase($post, $base_pen) if $post;
+	$rb->goto($line, 0);
+	my ($pre, undef, $post) = align textwidth($self->loading_message // ''), $self->body_cols // 1, 0.5;
+	$rb->erase($pre, $base_pen) if $pre;
+	$rb->text($self->loading_message, $base_pen);
+	$rb->erase($post, $base_pen) if $post;
 }
 
 =head2 on_scroll
@@ -945,7 +956,7 @@ Helper method to apply a series of coderefs to a value.
 sub fold_future {
 	my ($self, $prefix, $item, @steps) = @_;
 	return Future->wrap($item) unless @steps;
-	repeat {
+	try_repeat {
 		my $code = shift;
 		Future->call(sub { $code->(@$prefix, $item) })->on_done(sub {
 			$item = shift
@@ -970,20 +981,25 @@ Row cache accessor.
 sub row_cache {
 	my ($self, $row) = @_;
 	$self->{row_cache}[$self->row_cache_idx($row)] ||= do {
-		$self->adapter->get(
-			items => [$row],
-		)->then(sub {
-			# We have an item from storage. No idea what it is, could be an
-			# object, hashref, arrayref... the item transformations will
-			# convert it into something usable
-			my ($item) = @{ $_[0] };
-			return Future->fail('no such element', table => not_found => $row) unless $item;
+		my $found;
+		$self->adapter->range(
+			start => $row,
+			count => 1,
+			on_item => sub {
+				# We have an item from storage. No idea what it is, could be an
+				# object, hashref, arrayref... the item transformations will
+				# convert it into something usable
+				my ($idx, $item) = @_;
 
-			# Somewhat tedious way to reduce() a Future chain
-			$self->fold_future([ $row ], $item, @{$self->{item_transformations}})
+				# Somewhat tedious way to reduce() a Future chain
+				$found = $self->fold_future([ $row ], $item, @{$self->{item_transformations}})
+			}
+		)->then(sub {
+			$found || Future->done
 		})->then(sub {
 			# Our item is now accessible as an arrayref, start working on the columns
-			my $item = shift;
+			return Future->done unless @_;
+			my ($item) = @_;
 			my @pending;
 			for my $col (0..$#{$self->{columns}}) {
 				my $cell = $item->[$col];
@@ -1649,7 +1665,7 @@ And these are probably important background reading for formatting and data sour
 
 =head1 AUTHOR
 
-Tom Molesworth <cpan@perlsite.co.uk>
+Tom Molesworth <TEAM@cpan.org>
 
 =head1 CONTRIBUTORS
 
